from clusterenv.common import NodeGenerator, JobGenerator
from clusterenv.render import ClusterRenderer
from clusterenv._types import Jobs, Nodes, JobStatus
from gymnasium.core import ActType, ObsType, RenderFrame
from typing import List, Any, SupportsFloat, overload
import gymnasium as gym
import numpy as np
import logging


class ClusterEnv(gym.Env):

    metadata = {"render_modes": ["human", "rgb_array", ''], "render_fps": 4}

    def _action_shape(self) -> gym.spaces.Discrete:
        return gym.spaces.Discrete(len(self.nodes) * len(self.jobs) + 1)

    def _observation_shape(self) -> gym.spaces.Dict:
        max_val: float = np.max(self.nodes.original)
        return gym.spaces.Dict(
            dict(
                Usage=gym.spaces.Box(
                    low=0, high=max_val, shape=self.jobs.usage.shape, dtype=float
                ),
                Queue=gym.spaces.Box(
                    low=-1,
                    high=max_val,
                    shape=self.jobs.usage.shape,
                    dtype=float,
                ),
                Nodes=gym.spaces.Box(
                    low=0, high=max_val, shape=self.nodes.original.shape, dtype=float
                ),
                Status=gym.spaces.Box(
                    low=0, high=5, shape=self.jobs.status.shape, dtype=float
                ),
            )
        )

    def __init__(
        self,
        n_nodes: int,
        n_jobs: int,
        n_resource: int,
        max_time: int,
        node_gen: NodeGenerator = NodeGenerator,
        job_gen: JobGenerator = JobGenerator,
        render_mode: str = '',
        cooldown: float = 0.05
    ):
        super(ClusterEnv, self).__init__()
        self._time = 0
        self._mapper = np.arange(n_jobs)
        self.render_mode = render_mode
        self._n_resource = n_resource
        self._max_time = max_time
        self._n_nodes, self._n_jobs = n_nodes, n_jobs
        self._node_gen, self._job_gen = node_gen, job_gen
        self._logger = logging.getLogger(self.__class__.__name__)
        self._plotter = ClusterRenderer(
            nodes=self._n_nodes,
            jobs=self._n_jobs,
            resource=self._n_resource,
            time=self._max_time,
            cooldown=cooldown,
        )
        self._action_error = None
        self._action_correct = None
        self.nodes: Nodes = self._node_gen.generate(
            self._n_nodes, self._n_resource, self._max_time, 255.0
        )
        self.jobs: Jobs = self._job_gen.generate(
            self._n_jobs, self._n_resource, self._max_time, 255.0
        )
        self.observation_space = self._observation_shape()
        self.action_space = self._action_shape()

    def _reorganize_observation(self) -> None:
        self._mapper = self._mapper[np.argsort(self.jobs.status[self._mapper], kind='stable')]
        self.jobs.reorganize(self._mapper)
        # TODO: Update jobs
        self._observation = dict(
            Usage=self.nodes.usage,
            Nodes=self.nodes.original,
            Status=self.jobs.status,
            Queue=self.jobs.usage,
        )

    def _tick(self) -> None:
        """Forward cluster time by tick.
        Updating running time of all running jobs by one, show new income jobs.
        """
        self._logger.debug(f"Tick {self._time}->{self._time + 1}")
        self.jobs.inc(JobStatus.RUNNING)
        self._time += 1
        complete_jobs_idx: np.array = self.jobs.check_complete_jobs()
        arrived_jobs_idx: np.array = self.jobs.check_arrived_jobs(self._time)
        self._logger.debug(f"Receive new jobs: {arrived_jobs_idx}")
        self._logger.debug(f"Complete jobs: {complete_jobs_idx}")
        self.nodes.tick()

    def _schedule(self, n_idx: int, j_idx: int) -> bool:
        job_status, _, job_usage = self.jobs[j_idx]
        match job_status:
            case JobStatus.PENDING:
                free_n_space: np.ndarray = self.nodes.free_space(n_idx)
                if job_can_be_schedule := bool(np.all(free_n_space >= job_usage)):
                    self.nodes[n_idx] += job_usage
                    self.jobs[j_idx] = self.jobs[j_idx].change_status(JobStatus.RUNNING)
                    logging.debug(f"Succeed Allocated j.{j_idx} to n.{n_idx}")
                else:
                    logging.debug(f"Can't Allocate j.{j_idx} n.{n_idx}, not enough resource")
                return job_can_be_schedule
            case _:
                logging.debug(f"Can't Allocate j.{j_idx} with status {JobStatus(job_status).name}")
                return False

    @overload
    def _convert_action(self, n_idx: int, j_idx: int) -> int: ...

    @overload
    def _convert_action(self, action: int) -> tuple[int, int]: ...

    def _convert_action(self, *args):
        n_nodes: int = len(self.nodes)
        if len(args) == 2 and all(isinstance(arg, int) for arg in args):
            n_idx, j_idx = args
            return j_idx * n_nodes + n_idx
        elif len(args) == 1 and isinstance(args[0], int):
            action = args[0]
            return action % n_nodes, action // n_nodes
        else:
            raise TypeError("Invalid arguments for convert_action method")

    def _has_terminate(self):
        return len(self.jobs.index_by_status(JobStatus.COMPLETE)) == len(self.jobs)

    def _has_additional_jobs(self):
        n_pending: int = len(self.jobs.index_by_status(JobStatus.COMPLETE))
        n_complete: int = len(self.jobs.index_by_status(JobStatus.RUNNING))
        return n_pending + n_complete != len(self.jobs)

    def _total_wait_time(self):
        return self.jobs.total_pending_time()

    def render(self) -> RenderFrame | List[RenderFrame] | None:
        if self.render_mode:
            fig = self._plotter(
                self._observation,
                current_time=self._time,
                error=self._action_error,
                correct=self._action_correct,
            )
            if self.render_mode == "rgb_array":
                buf = fig.canvas.tostring_rgb()
                width, height = fig.canvas.get_width_height()
                expected_height = int(fig.get_figheight() * fig.dpi)
                expected_width = int(fig.get_figwidth() * fig.dpi)
                width_mult: int = expected_width // width
                height_mult: int = expected_height // height
                return np.frombuffer(buf, dtype=np.uint8).reshape(
                    (height_mult * height, width_mult * width, 3)
                )

    def reset(
            self, *, seed: int | None = None, options: dict[str, Any] | None = None
    ) -> tuple[Any, dict[str, Any]]:
        self.nodes: Nodes = self._node_gen.generate(
            self._n_nodes, self._n_resource, self._max_time, 255.0
        )
        self.jobs: Jobs = self._job_gen.generate(
            self._n_jobs, self._n_resource, self._max_time, 255.0
        )
        self.jobs.check_arrived_jobs(self._time)
        self._reorganize_observation()
        self.render()
        return self._observation, {}

    def step(
        self, action: ActType
    ) -> tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        """
        Make Cluster Step. Each step is a sub-time action, in other words player can pick inf action and time will not tick.
        Time will only tick when player take time tick action. Any other action is a scheduling action.
        When player take not possible action an error flag of self._action_error will be activated for wrapper Env (None).

        :param action: flatten action
        """
        info: dict = {}
        self._action_error = None
        self._action_correct = None
        self.render()
        if bool(action == 0):
            self._logger.debug(f"Tick Cluster ...")
            self._tick()
        else:
            n_idx, j_idx = self._convert_action(int(action) - 1)
            if not self._schedule(n_idx, j_idx):
                self._action_error = (n_idx, j_idx)
            else:
                self._action_correct = (n_idx, j_idx)
        terminate: bool = self._has_terminate()
        truncated: bool = not self._has_additional_jobs()
        reward: float = -self._total_wait_time()
        self._reorganize_observation()
        self.render()
        return self._observation, reward, terminate, truncated, info



from typing import NamedTuple, MutableSequence, TypeVar, Sequence, overload, Optional
from dataclasses import dataclass, field
from abc import abstractmethod
import numpy.typing as npt
from enum import IntEnum
import numpy as np


T = TypeVar("T", bound="JobStruct")
_T = TypeVar("_T")


class NodeStatus(IntEnum):
    DOWN = 0
    RESTARTING = 1
    UP = 2


class JobStatus(IntEnum):
    PENDING = 0
    NONEXISTENT = 1
    RUNNING = 2
    COMPLETE = 3


class JobStruct(NamedTuple):
    status: int
    submission: int
    usage: npt.NDArray[float]

    def change_status(self, status: JobStatus) -> "JobStruct":
        return JobStruct(status, submission=self.submission, usage=self.usage)


class Nodes(Sequence[_T]):
    __slots__ = ("_original", "_usage")
    _original: npt.NDArray[float]
    _usage: npt.NDArray[float]

    @property
    def original(self) -> npt.NDArray[float]:
        return self._original

    @property
    def usage(self) -> npt.NDArray[float]:
        return self._usage

    def __init__(self, nodes: npt.NDArray[float]):
        self._original = nodes
        self._usage = np.zeros(self._original.shape)

    def tick(self):
        self._usage = np.roll(self._usage, shift=-1, axis=-1)
        self._usage[:, :, -1] = 0

    def free_space(self, index: int) -> npt.NDArray[float]:
        return self._original[index] - self._usage[index]

    def __getitem__(self, index: int) -> npt.NDArray[float]:
        return self._usage[index]

    def __setitem__(self, index: int, value: npt.NDArray[float]) -> None:
        self._usage[index] = value

    def __len__(self):
        return self._usage.shape[0]


@dataclass
class Jobs(Sequence[T]):
    EMPTY_JOB_USAGE = -1
    usage: npt.NDArray[np.float64]
    submission: npt.NDArray[int]
    _status: npt.NDArray[int] = field(init=False)
    _length: npt.NDArray[int] = field(init=False)
    _run_time: npt.NDArray[int] = field(init=False)
    _wait_time: npt.NDArray[int] = field(init=False)

    @property
    def status(self) -> npt.NDArray[int]:
        return self._status

    def __post_init__(self):
        _job_start_time: npt.NDArray[np.uint32] = (self.usage == 0).argmax(axis=-1)
        self._status = np.full(
            shape=len(self.submission),
            fill_value=JobStatus.NONEXISTENT.value,
            dtype=np.uint32,
        )
        self._length = np.max(
            np.where(_job_start_time > 0, _job_start_time, self.usage.shape[0]), axis=1
        ).astype(np.uint32)
        self._run_time = np.zeros(len(self)).astype(int)
        self._wait_time = np.zeros(len(self)).astype(int)

    def reorganize(self, mapping: npt.ArrayLike) -> None:
        self.submission = self.submission[mapping]
        self.usage = self.usage[mapping]
        self._status = self._status[mapping]
        self._wait_time = self._wait_time[mapping]
        self._run_time = self._run_time[mapping]
        self._length = self._length[mapping]

    def total_pending_time(self) -> int:
        return int(np.sum(self._wait_time))

    def index_by_status(self, status: JobStatus) -> npt.NDArray[int]:
        return np.array(np.where(self._status == status))[0]

    def inc(self, status: JobStatus) -> None:
        job_indexes: npt.NDArray[int] = np.array(np.where(self._status == status))
        match status:
            case JobStatus.RUNNING:
                self._run_time[job_indexes] += 1
            case JobStatus.PENDING:
                self._wait_time[job_indexes] += 1
            case JobStatus.NONEXISTENT | JobStatus.COMPLETE:
                raise ValueError

    def check_complete_jobs(self) -> npt.NDArray[int]:
        complete: npt.NDArray[int] = np.array(np.where(self._run_time == self._length))
        self._status[complete] = JobStatus.COMPLETE
        self.usage[complete] = self.EMPTY_JOB_USAGE
        return complete

    def check_arrived_jobs(self, current_time: int) -> npt.NDArray[int]:
        arrived: npt.NDArray[int] = np.array(np.where(self.submission == current_time))
        self._status[arrived] = JobStatus.PENDING

        return arrived

    def __getitem__(self, index: int) -> T:
        return JobStruct(
            status=self._status[index],
            usage=self.usage[index].copy(),
            submission=self.submission[index],
        )

    def __setitem__(self, index: int, value: T) -> None:
        self._status[index] = value.status
        self.usage[index] = value.usage.copy()
        self.submission[index] = value.submission

    def __delitem__(self, index: int) -> None:
        raise ValueError(f"Can't Delete form this object: {type(self)}")

    def __len__(self):
        return len(self._status)



from numba import float64, int32, njit, jit, deferred_type
from numba.experimental import jitclass
import numpy as np
import gymnasium as gym
from clusterenv._types import JobStatus
from typing import Protocol, Dict
from clusterenv._types import Nodes, Jobs
import logging


@njit
def _node_generate(n: int, r: int, t: int, max_usage: float) -> np.ndarray:
    nodes: np.array = np.full(
        (n, r, t),
        fill_value=max_usage,
        dtype=float,
    )
    return nodes


def _job_generate(n: int, r: int, t: int, max_usage: float, arrival: Dict[str, float], length: Dict[str, float], usage: Dict[str, float]) -> tuple[np.ndarray, np.array]:
    submission: np.array = t * np.random.choice(
        arrival["option"], size=n, p=arrival["prob"]
    ).astype(int)
    length: np.array = 1 + t * np.random.choice(
        length["option"], size=(n, r), p=length["prob"]
    )
    usage: np.array = max_usage * np.random.choice(
        usage["option"], size=n, p=usage["prob"]
    ).astype(float)
    usage: np.ndarray = np.tile(usage[..., np.newaxis, np.newaxis], (r, t))
    mask = np.arange(usage.shape[-1]) >= length[..., np.newaxis]
    usage[mask] = 0.0
    return usage, submission


class NodeGenerator(Protocol):

    @classmethod
    def generate(cls, n: int, r: int, t: int, max_usage: float) -> Nodes:
        return Nodes(_node_generate(n, r, t, max_usage))


class JobGenerator(Protocol):
    LENGTH = dict(option=[0.0, 0.2, 0.3], prob=[0.7, 0.2, 0.1])
    USAGE = dict(option=[1], prob=[1])
    ARRIVAL_TIME = dict(option=[0.0], prob=[1])

    @classmethod
    def generate(cls, n: int, r: int, t: int, max_usage: float) -> Jobs:
        return Jobs(*_job_generate(n, r, t, max_usage, length=cls.LENGTH, arrival=cls.ARRIVAL_TIME, usage=cls.USAGE))


spec = [
    ('nodes', float64[:, :, :]),
    ('jobs', float64[:, :, :]),
    ('_logger', float64),
    ('_run_time', float64[:]),
    ('_max_val', float64),
    ('_time', int32)
]

@jitclass(spec)
class Cluster:

    def __init__(self, nodes, jobs):
        self.nodes = nodes
        self.jobs = jobs
        self._logger = logging.getLogger(self.__class__.__name__)
        self._run_time = np.zeros(len(self.jobs)).astype(float)
        self._max_val = np.max(self.nodes)
        self.jobs.check_arrived_jobs(self._time)

    @property
    def observation(self):
        return dict(
            Usage=self.nodes.usage,
            Queue=self.jobs.usage,
            Nodes=self.nodes.original,
            Status=self.jobs.status,
        )

    @property
    def time(self):
        return self._time

    @property
    def max_usage(self):
        return self._max_val

    def action_space(self):
        return gym.spaces.Discrete(len(self.nodes) * len(self.jobs) + 1)

    def observation_shape(self):
        max_val = np.max(self.nodes.original)
        return gym.spaces.Dict(
            dict(
                Usage=gym.spaces.Box(
                    low=0, high=max_val, shape=self.jobs.usage.shape, dtype=float
                ),
                Queue=gym.spaces.Box(
                    low=-1,
                    high=max_val,
                    shape=self.jobs.usage.shape,
                    dtype=float,
                ),
                Nodes=gym.spaces.Box(
                    low=0, high=max_val, shape=self.nodes.original.shape, dtype=float
                ),
                Status=gym.spaces.Box(
                    low=0, high=5, shape=self.jobs.status.shape, dtype=float
                ),
            )
        )

    def has_terminate(self):
        return len(self.jobs.index_by_status(JobStatus.COMPLETE)) == len(self.jobs)

    def has_additional_jobs(self):
        n_pending = len(self.jobs.index_by_status(JobStatus.COMPLETE))
        n_complete = len(self.jobs.index_by_status(JobStatus.RUNNING))
        return n_pending + n_complete != len(self.jobs)

    def total_wait_time(self):
        return self.jobs.total_pending_time()

    def tick(self):
        self._logger.info(f"Tick {self._time}->{self._time+1}")
        self.jobs.inc(JobStatus.RUNNING)
        self._time += 1
        complete_jobs_idx = self.jobs.check_complete_jobs()
        arrived_jobs_idx = self.jobs.check_arrived_jobs(self._time)
        self._logger.info(f"Receive new jobs: {arrived_jobs_idx}")
        self._logger.info(f"Complete jobs: {complete_jobs_idx}")
        self.nodes.tick()

    def schedule(self, n_idx, j_idx):
        job_status, _, job_usage = self.jobs[j_idx]
        if job_status == JobStatus.PENDING:
            free_n_space = self.nodes.free_space(n_idx)
            job_can_be_schedule = np.all(free_n_space >= job_usage)
            if job_can_be_schedule:
                self.nodes[n_idx] += job_usage
                self.jobs[j_idx] = self.jobs[j_idx].change_status(JobStatus.RUNNING)
                logging.debug(f"Succeed Allocated j.{j_idx} to n.{n_idx}")
            else:
                logging.debug(f"Can't Allocate j.{j_idx} n.{n_idx}, not enough resource")
            return job_can_be_schedule
        else:
            logging.debug(f"Can't Allocate j.{j_idx} with status {JobStatus(job_status).name}")
            return False



from clusterenv._types import JobStatus
from typing import Iterable, Callable, Any
from dataclasses import dataclass, field
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
import numpy.typing as npt
import numpy as np
import math

from numba import njit, jit

CMAP_COLORS: tuple = ("copper", "gray", "twilight", "summer")
REGULAR_COLOR: str = "copper"
REGULAR_TITLE_COLOR: str = "black"
ERROR_COLOR: str = "RdGy"
CORRECT_COLOR: str = "Greens"
ERROR_TITLE_COLOR: str = "red"


def _hide_unused(axs: Axes,  nodes: np.array, jobs: np.array, nodes_n_columns: int):
    nodes_to_remove: list = axs[:, :nodes_n_columns].flatten()[nodes:]
    jobs_to_remove: list = axs[:, nodes_n_columns:].flatten()[jobs:]
    for ax in nodes_to_remove:
        plt.delaxes(ax)
    for ax in jobs_to_remove:
        plt.delaxes(ax)


@jit(forceobj=True)
def _draw(
    matrix: np.ndarray,
    /,
    *,
    title: str,
    title_color: str,
    ax: Axes,
    time: int,
    resource: int,
    cmap: str,
):
    ax.imshow(matrix, cmap=cmap, vmin=0, vmax=100)
    ax.set_title(title, fontsize=10, color=title_color)
    ax.set_xticks(np.arange(0, time, 0.5), minor=True)
    ax.set_yticks(np.arange(0, resource, 0.5), minor=True)
    ax.tick_params(which="minor", length=0)
    ax.grid(which="both", color="black", linestyle="-", linewidth=0.5, alpha=0.3)
    ax.set_xticks([])
    ax.set_yticks([])


@jit(forceobj=True)
def _draw_job(
    job: np.ndarray,
    /,
    *,
    title_color: str,
    idx: int,
    ax: Axes,
    time: int,
    resource: int,
    cmap: str,
    status: JobStatus,
):
    title: str = f"[J.{idx}]"
    cmap = cmap if cmap == ERROR_COLOR else CMAP_COLORS[status]
    _draw(
        job,
        title=title,
        title_color=title_color,
        ax=ax,
        time=time,
        resource=resource,
        cmap=cmap,
    )


@jit(forceobj=True)
def _draw_node(
    node: np.ndarray,
    /,
    *,
    title_color: str,
    idx: int,
    ax: Axes,
    time: int,
    resource: int,
    cmap: str,
):
    title: str = f"[N.{idx}]"
    _draw(
        node,
        title=title,
        title_color=title_color,
        ax=ax,
        time=time,
        resource=resource,
        cmap=cmap,
    )


def color_map(error, correct):
    def inner(idx, pos):
        if error and idx == error[pos]:
            return ERROR_COLOR
        elif correct and idx == correct[pos]:
            return CORRECT_COLOR
        return REGULAR_COLOR
    return inner


@dataclass(slots=False)
class ClusterRenderer:
    """
    Craete Gym Cluster Object.
    Allow To Represent Cluster logic: Scheduling, TimeTick, Generating Job.
    Job & Node reperesentation should be similar.

    :param nodes: List of nodes repersentation
    :param jobs: List of job representation
    :param resource: Number of Job/Node resource type
    :param max_time: Max Job Run time
    :param cooldown: Render cooldown between stepes
    """

    nodes: int
    jobs: int
    resource: int
    time: int
    cooldown: float
    fig: Figure = field(init=False)
    axs: npt.NDArray = field(init=False)

    def __post_init__(self):
        self.jobs_n_columns: int = math.ceil(self.jobs**0.5)
        self.nodes_n_columns: int = math.ceil(self.nodes**0.5)

        jobs_n_rows: int = math.ceil(self.jobs / self.jobs_n_columns)
        nodes_n_rows: int = math.ceil(self.nodes / self.nodes_n_columns)

        n_rows: int = max(jobs_n_rows, nodes_n_rows)
        n_columns: int = self.nodes_n_columns + self.jobs_n_columns

        self.fig, self.axs = plt.subplots(
            n_rows, n_columns, figsize=(8, 4), facecolor="white"
        )
        _hide_unused(
            self.axs,
            nodes=self.nodes,
            jobs=self.jobs,
            nodes_n_columns=self.nodes_n_columns,
        )

    def __call__(
        self,
        obs: dict[str, np.ndarray],
        /,
        *,
        current_time: int,
        error: None | tuple[int, int],
        correct: None | tuple[int, int]
    ) -> Any:
        self.fig.suptitle(f"Time: {current_time}", fontsize=16, fontweight="bold")
        nodes: npt.NDArray = obs["Usage"]
        queue: npt.NDArray = obs["Queue"]
        status: npt.NDArray[np.uint32] = obs["Status"]
        cmap_color = color_map(error, correct)
        title_color: Callable[[int, int], str] = (
            lambda idx, pos: ERROR_TITLE_COLOR
            if error and idx == error[pos]
            else REGULAR_TITLE_COLOR
        )
        node_ax: Callable[[int], npt.NDArray] = lambda n_idx: self.axs[
            n_idx // self.nodes_n_columns, n_idx % self.nodes_n_columns
        ]
        job_ax: Callable[[int], npt.NDArray] = lambda j_idx: self.axs[
            j_idx // self.jobs_n_columns,
            self.nodes_n_columns + (j_idx % self.jobs_n_columns),
        ]
        # update matries
        for n_idx, node in enumerate(nodes):
            _draw_node(
                node,
                title_color=title_color(n_idx, 0),
                idx=n_idx,
                ax=node_ax(n_idx),
                time=self.time,
                resource=self.resource,
                cmap=cmap_color(n_idx, 0),
            )
        for j_idx, job in enumerate(queue):
            _draw_job(
                job,
                title_color=title_color(j_idx, 1),
                idx=j_idx,
                ax=job_ax(j_idx),
                time=self.time,
                resource=self.resource,
                cmap=cmap_color(j_idx, 1),
                status=JobStatus(status[j_idx]),
            )
        # update figure
        plt.draw()
        plt.pause(self.cooldown)
        return self.fig


from gymnasium.envs.registration import register

register(
    id="cluster-v0",
    entry_point="clusterenv.envs.cluster:ClusterEnv",
    kwargs=dict(
        n_nodes=5,
        n_jobs=10,
        n_resource=3,
        max_time=5,
    ),
)
